import logging
import os
import json
import re
from string import punctuation as PUNCTUATION

from dataclasses import dataclass
from thefuzz import process as fuzz_process

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

##TODO add broadway edge case


class AddressError(Exception):
    """Base Exception for errors related to the Address class"""

    pass


class AddressLengthError(AddressError):
    """Error raised when an entered address does not contain at least 3 elements"""

    pass


class NonIntegerStreetNumberError(AddressError):
    """Error raise when an entered street number is not an integer."""


class NoCloseMatchError(Exception):
    """Error raised when no close match is found in match_closest_street_name"""


class Address:
    """Dataclass to represent address. str returns the street address as a string and
    len the number of elements in the street address."""

    street_number: str
    street_name: str

    def __init__(self, street_address: str, raise_invalid_error: bool = True):
        street_address_list = street_address.split(" ")

        self.street_number = street_address_list[0]

        self.street_name = " ".join(street_address_list[1:])

        if raise_invalid_error:
            self.is_valid_address(raise_error=True)

    def __str__(self):
        return self.street_address

    def __len__(self):
        return len(self.get_address_as_list())

    def get_address_as_list(self):
        """Converts the inputted address to a list if it is in string form. Returns the address as a list"""
        address_list = self.street_address.split(" ")
        return address_list

    @property
    def street_type(self):
        return self.street_name.split(" ")[-1]

    @property
    def street_address(self):
        return f"{self.street_number} {self.street_name}"

    def check_address_arg_length(self) -> bool:
        """Returns false if the input address doesn't have at least 3 components"""
        arg_length = len(self)
        if arg_length < 3:
            # Address should contain {number} {street_name} {street_suffix} at minimum
            raise AddressLengthError(
                f"Street address {self.street_address} does not contain enough elements (number name type)"
            )
        else:
            return True

    def check_street_number(self) -> bool:
        """Returns True if the street number is valid"""
        # ensure first input is an integer
        try:
            # should throw value error if street number cannot be returned
            int(self.street_number)
            return True
        except ValueError:
            raise NonIntegerStreetNumberError(
                f"{self.street_number} is not a valid numeric street number"
            )

    def format_numeric_street_name(self):
        """adds '0' to street name if street name is numeric, <10, and doesn't contain '0' (ex 9th st -> 09th st)"""
        street_name = self.street_name
        if street_name[0].isnumeric() and not street_name[1].isnumeric():
            new_street_name = "0" + street_name
            self.street_name = new_street_name
        return True

    def is_valid_address(self, raise_error: bool = False):
        """Verifies an the Address object is a valid street address and returns true if it is.
        If raise_error is false(default), an invalid address will return False. If raise_error is passed as True,
        the method will raise the corresponding error."""

        try:
            self.check_address_arg_length()
        except AddressLengthError as err:
            if raise_error:
                raise AddressLengthError(err)
            else:
                logging.warning(f"Excepted error: {err}")
                return False

        try:
            self.check_street_number()
        except NonIntegerStreetNumberError as err:
            if raise_error:
                raise NonIntegerStreetNumberError(err)
            else:
                logging.warning(f"Excepted error: {err}")
                return False

        return True


def load_street_type_abbreviations(
    path: str = os.path.join(DATA_DIR, "street_types.json")
) -> dict:
    """Loads the street type dictionary json file and returns it as a dict. Raises FileNotFoundError if the file is not found."""
    try:
        with open(path, "r") as fp:
            street_type_dict = json.load(fp)
            return street_type_dict
    except FileNotFoundError:
        raise FileNotFoundError(
            f"street_type_dict_file not found at {os.path.abspath(path)}"
        )


def remove_city(address_input: str) -> str:
    """Removes the city/state/zipcode from an inputted string address and returns only the street address. If there
    is no city/state/zipcode info, the original string is returned. *only works for San Francisco*"""

    address_list = address_input.split(",")

    if len(address_list) > 1:
        if Address(address_list[0], raise_invalid_error=False).is_valid_address():
            return address_list[0]

        else:
            address_input = address_input.replace(",", "")

    search = re.search("San Francisco", address_input, re.IGNORECASE)

    if search:
        address = address_input.split(search.group())[0]
        address = address.strip()
    else:
        address = address_input

    return address


def abbreviate_street_type(address: Address) -> Address:
    """Abbreviates the street type if it is in it's long form (i.e. street -> st). Returns a new Address object.
    If the address cannot or does not need to be abbreviated, the original Address is returned."""

    # load the street_type_dictionaries
    try:
        street_abbreviation_dict = load_street_type_abbreviations()
    except FileNotFoundError as err:
        logging.error(f"{err}: street not abbreviated.")
        return address

    street_type = address.street_type.lower()

    if street_type in street_abbreviation_dict.keys():
        # get the corressponding abbreviated street type and create a new street name string with the long type replaced
        abbreviated_street_type = street_abbreviation_dict[street_type]
        new_street_name = address.street_name.replace(
            address.street_type, abbreviated_street_type
        )
        # create a new Address
        abbreviated_address = Address(f"{address.street_number} {new_street_name}")
        return abbreviated_address

    # if the street type is not found in the dict, return the original address
    else:
        return address


def remove_punctuation(input: str, *exceptions: str) -> str:
    """Removes all PUNCTUATION from a string and returns the stripped string.
    Excepts exceptions from PUNCTUATION removal."""
    allowed_punctuation = PUNCTUATION
    for exception in exceptions:
        allowed_punctuation = allowed_punctuation.replace(exception, "")

    allowed_punctuation_set = re.compile(f"[{re.escape(allowed_punctuation)}]")
    return allowed_punctuation_set.sub("", input)


def create_standard_Address(user_input: str) -> Address:
    """Takes a user string input and creates a 'standard' Address object.
    This entails:
        - Street address only
        - Lower case
        - No random puncuation
        - Abbreviated street type
    Returns the standard Address. Will raise corresponding errors if users input is invalid."""

    address_string = user_input.lower().strip()
    address_string = remove_city(address_string)
    address_string = remove_punctuation(address_string, "-")
    address = Address(address_string)
    address = abbreviate_street_type(address)
    return address


def load_street_names(path: str = os.path.join(DATA_DIR, "street_names.json")) -> list:
    """Loads and returns list of street names in SF Tree Database."""
    try:
        with open(path, "r") as fp:
            street_names = json.load(fp)
            return street_names
    except FileNotFoundError:
        raise FileNotFoundError(
            f"street_names file not found at {os.path.abspath(path)}"
        )


def match_closest_street_name(
    address: Address, streets: list[str], min_score: int = 90
) -> Address:
    """Match the Address objects street name to a queryable street name in streets (from SF_Trees.db).
    If perfect match, the original Address will be returned.
    If there is a match with a score greater than min_score, a new Address object with that matched street name will be returned.
    If no close match is found, NoCloseMatchError is raised."""

    street_name = address.street_name

    if street_name in streets:
        return address

    closest_match, score = fuzz_process.extractOne(
        street_name, streets
    )  # specify method
    if score > min_score:
        address.street_name = closest_match
        return address
    raise NoCloseMatchError(
        f"Street name match for {address.street_name} doesn't meet minimum score of {min_score}. "
        f"{closest_match=} {score=}"
    )


def get_Address_for_query(user_input: str) -> Address:
    address = create_standard_Address(user_input)
    street_names = load_street_names()
    return match_closest_street_name(address, street_names)



import os.path
import logging
import requests
from requests.exceptions import HTTPError
import time
from dataclasses import dataclass
import pandas as pd
import numpy as np
from thefuzz import fuzz
import random
import argparse

# TODO add in X addresses

"""Search results from selec tree API are in the form of:
{
  "totalResults": 82,
  "pageResults": [
    {
      "tree_id": 39,
      "family": "Sapindaceae",
      "common": "TRIDENT MAPLE",
      "name_concat": "<em>Acer</em> <em>buergerianum</em>",
      "name_unformatted": "Acer buergerianum",
      "link": "/images/0000/39/original/acer-buergerianum-acer-buergerianum-ro.jpg",
      "sequence": 1,
      "low_zone": 1,
      "height_high": 25
    },
  "sort": 1,
  "resultsPerPage": "30",
  "activePage": "1"
}
"""

global score_count
score_count = [0, 0]  # total score, number of times score is calculated

FILENAME = "mapped_species"
logging.basicConfig(filename=f"{FILENAME}.log", filemode="w", level=logging.WARNING)


class SelecTreeResultNotFoundError(Exception):
    """Raised when a SelecTree result isn't found."""

    pass


@dataclass
class Specie:
    scientific_name: str
    common_name: str
    tree_id: int

    def __init__(
        self,
        scientific_name=None,
        common_name=None,
        formatted_name=None,
        pageResult=None,
        tree_id=None,
    ):
        if (
            (common_name is None or scientific_name is None)
            ^ (formatted_name is None)
            ^ (pageResult is None)
        ):
            raise ValueError("Too many args passed to init Specie")

        elif pageResult:
            (
                self.tree_id,
                self.scientific_name,
                self.common_name,
            ) = self._set_Specie_from_pageResult(pageResult)

        elif formatted_name:
            (
                self.scientific_name,
                self.common_name,
            ) = self._set_Specie_from_formatted_name(formatted_name)

        else:
            self.scientific_name = scientific_name
            self.common_name = common_name

        if not hasattr(self, "tree_id"):
            self.tree_id = tree_id

    @property
    def formatted_name(self) -> str:
        if self.common_name and self.scientific_name:
            return f"{self.scientific_name} :: {self.common_name}"

        elif self.scientific_name:
            return f"{self.scientific_name} ::"

        elif self.common_name:
            return f":: {self.common_name}"

        else:
            raise UserWarning("No common or scientific name set for Specie")
            return None

    @property
    def full_name(self):
        if self.common_name and self.scientific_name:
            return f"{self.scientific_name} {self.common_name}"

        elif self.scientific_name:
            return f"{self.scientific_name}"

        elif self.common_name:
            return f"{self.common_name}"

        else:
            raise UserWarning(f"No common or scientific name set for {self}")
            return None

    def _set_Specie_from_formatted_name(self, species_name: str):
        """ "Returns the scientific and common name from the species.csv qSpecies. If no common name exists, the common
        name atttribute will be assigned None."""
        if species_name.endswith(" ::"):
            scientific_name = species_name.replace(" ::", "")
            return scientific_name, None
        else:
            scientific_name, common_name = species_name.split(" :: ")
            return scientific_name, common_name

    def _set_Specie_from_pageResult(self, pageResult: dict):
        """Creates a new Specie object from a selec tree pageResult dict passed as Specie(pageResult={pageResult dict}.
        Will assign tree_id from pageResult to new Specie object."""
        tree_id = pageResult["tree_id"]
        common_name = pageResult["common"]
        scientific_name = (
            pageResult["name_unformatted"].replace("<em>", "").replace("</em>", "")
        )
        return tree_id, scientific_name, common_name


def _get_match_scores(
    specie: Specie, search_results: list[Specie], property: str
) -> np.array:
    """Base function for assigning matching scores between Specie specie to list of other Species.
    The variable 'property' determines which (string) property of the Specis is being matched between.
    Returns a numpy array of the scores with the same indices as the original search_result list passed to the function."""
    scores = np.zeros(len(search_results))
    for i, result in enumerate(search_results):
        score = fuzz.partial_token_sort_ratio(
            getattr(specie, property).lower(), getattr(result, property).lower()
        )
        scores[i] = score

    return scores


def get_common_name_match_scores(specie: Specie, search_results: list[Specie]):
    """Implements _get_match_scores() for the common_name Species property"""
    if specie.common_name:
        return _get_match_scores(specie, search_results, "common_name")
    else:
        raise ValueError(f"No common name assigned to {specie}")


def get_scientific_name_match_scores(specie: Specie, search_results: list[Specie]):
    """Implements _get_match_scores() for the common_name Species property"""
    if specie.scientific_name:
        return _get_match_scores(specie, search_results, "scientific_name")

    else:
        raise ValueError(f"No scientific name assigned to {specie}")


def query_selec_tree(search_term: str, results_per_page: int = 10) -> list:
    """Queries selec tree for the passed search term. Returns the results as a list of Species objects."""
    url = "https://selectree.calpoly.edu/api/search-by-name-multiresult"
    payload = {
        "searchTerm": search_term,
        "activePage": 1,
        "resultsPerPage": results_per_page,
        "sort": 1,
    }
    r = requests.get(url, params=payload, timeout=2)

    if r.status_code == 200:
        search_results = r.json()["pageResults"]

        search_result_species = list()
        for result in search_results:
            search_result_species.append(Specie(pageResult=result))

        return search_result_species

    try:
        r.raise_for_status()

    except HTTPError as err:
        raise HTTPError(f"Error for {search_term}: {err}")

    except Exception as err:
        raise ConnectionError(f"Connection error (non-HTTP) for {search_term}: {err}")


def find_closest_match(
    specie: Specie,
    possible_ids: list[Specie],
    key: str,
    weight: int,
    minimum_score: int = 55,
):
    num_results = len(possible_ids)

    if num_results == 1:
        result_specie = possible_ids[0]
        logging.warning(f"Result for {specie}: {result_specie.formatted_name}")

        return result_specie.tree_id

    # full_name vs common_name vs scientific_name logic
    if key == "full_name":
        scientific_name_scores = get_scientific_name_match_scores(specie, possible_ids)
        if max(scientific_name_scores) == 100:  # perfect match
            location = np.where(scientific_name_scores == 100)[0][0]
            logging.warning(
                f"The scientific name was a perfect match for {specie}: {possible_ids[location]}"
            )
            return possible_ids[location].tree_id

        common_name_scores = get_common_name_match_scores(specie, possible_ids)
        if max(common_name_scores) == 100:  # perfect match
            location = np.where(common_name_scores == 100)[0][0]
            logging.warning(
                f"The common name was a perfect match for {specie}: {possible_ids[location]}"
            )
            return possible_ids[location].tree_id

        # weighted average of two scores
        total_scores = (scientific_name_scores * weight + common_name_scores) / (
            weight + 1
        )
        if max(total_scores) > minimum_score:
            location = np.where(total_scores == total_scores.max())[0][0]
            logging.warning(
                f"Search results using the {key} of {specie} returned {possible_ids[location]}. "
                f"Weighted score = {total_scores.max()}"
            )
            return possible_ids[location].tree_id

    elif key == "scientific_name":
        scientific_name_scores = get_scientific_name_match_scores(specie, possible_ids)
        if max(scientific_name_scores) == 100:  # perfect match
            location = np.where(scientific_name_scores == 100)[0][0]
            return possible_ids[location].tree_id

        if max(scientific_name_scores) > minimum_score:
            location = np.where(scientific_name_scores == scientific_name_scores.max())[
                0
            ][0]
            logging.warning(
                f"Search results using the {key} of {specie} returned {possible_ids[location]}. "
                f"Score = {scientific_name_scores.max()}"
            )
            return possible_ids[location].tree_id

    elif key == "common_name":
        common_name_scores = get_common_name_match_scores(specie, possible_ids)
        if max(common_name_scores) == 100:  # perfect match
            location = np.where(common_name_scores == 100)[0][0]
            logging.warning(
                f"The common name was a perfect match for {specie}: {possible_ids[location]}"
            )
            return possible_ids[location].tree_id

        if max(common_name_scores) > minimum_score:
            location = np.where(common_name_scores == common_name_scores.max())[0][0]
            logging.warning(
                f"Search results using the {key} of {specie} returned {possible_ids[location]}. "
                f"Score = {common_name_scores.max()}"
            )
            return possible_ids[location].tree_id

    else:
        raise Error("get_selec_tree_url_path key not found in matching statements")

    # if no id is returned
    logging.error(f"No strong match for {specie}")
    return 0


def get_selec_tree_url_path(specie: Specie, weight: float = 1) -> int:
    """Takes a tree specie Specie as argument and returns the url path from selec tree to the closest matching page."""
    if not specie.common_name:
        search_terms = {"scientific_name": specie.scientific_name}

    else:
        search_terms = {
            "full_name": specie.full_name,
            "scientific_name": specie.scientific_name,
            "common_name": specie.common_name,
        }

    num_results = 0

    for key in search_terms:
        search_term = search_terms[key]
        possible_ids = query_selec_tree(search_term)
        num_results = len(possible_ids)
        if num_results > 0:
            break
    else:
        raise SelecTreeResultNotFoundError

    return find_closest_match(specie, possible_ids, key, weight)


def assign_url_paths(
    species: pd.Series,
    time_buffer: bool = True,
    show_progress: bool = False,
    weight: float = 1.2,
) -> pd.DataFrame:
    """Takes the species series as input and returns a dataframe containing the original series
    and the url path number (key) appended as a coloumn"""

    def map_url(species_name, time_buffer: bool = True) -> int:
        # time buffer to not abuse Selec Tree or be banned
        if time_buffer:
            wait_time = random.randint(1, 3) * 2
            time.sleep(wait_time)

        id = get_selec_tree_url_path(species_name, weight=1.2)
        return id

    num_species = len(species)
    urlPaths = pd.Series(np.zeros(num_species, dtype="uint16"))

    for i, specie_name in species.items():
        try:
            specie = Specie(formatted_name=specie_name)
            urlPaths.loc[i] = map_url(specie, time_buffer)

        except HTTPError as err:
            logging.error(f"HTTP error while mapping {species[i]}: {err}")
            urlPaths.loc[i] = 0

        except ConnectionError as err:
            logging.error(f"Connection error while mapping {species[i]}: {err}")
            urlPaths.loc[i] = 0

        except SelecTreeResultNotFoundError:
            logging.error(f"Selec Tree result not found while mapping {species[i]}")
            urlPaths.loc[i] = 0

        except Exception as err:
            logging.error(f"Exception while mapping {species[i]}: {err=}")
            urlPaths.loc[i] = 0

        if show_progress:
            print(f"{i + 1}/{num_species}", end="\r")

            if i + 1 == num_species:
                # new line to overwrite carriage
                print()

    mapped_series = pd.concat([species, urlPaths], axis=1).rename(
        {0: "urlPath", "0": "qSpecies"}, axis=1
    )

    return mapped_series


if __name__ == "__main__":
    start_time = time.time()
    species_path = os.path.join("src", "SF_Tree_Identifier", "data", "Species.csv")
    species_series = pd.read_csv(species_path, index_col=0).iloc[:, 0]
    species_df = assign_url_paths(species_series, show_progress=True, weight=1.2)
    species_df.to_csv(f"{FILENAME}.csv")

    # calculate number missing vs complete
    num_missing = species_df.loc[species_df.urlPath == 0].urlPath.size
    total_species, _ = species_df.shape

    # calculate time taken
    time_taken = time.time() - start_time
    mins = int(time_taken // 60)
    seconds = time_taken % 60

    # print results
    print(f"Done! Time taken: {mins}min {seconds:.0f}s")
    print(f"Number of species missing paths: {num_missing}/{total_species}")
    # if score_count[1] > 0:
    # print(f'Multi result occured {score_count[1]} times, average score was {score_count[0] / score_count[1]: .0f}')



import logging
import sqlite3
import os
from timeit import timeit
from collections import namedtuple

import pandas as pd

import Address

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
DB_LOCATION = os.path.join(DATA_DIR, "SF_Trees.db")


def create_address_query(address: Address.Address) -> str:
    """Creates the sql query to retrieve the qSpecies keys for a specfic address. Returns the query as a string."""
    query = f"""
        SELECT qSpecies
        FROM addresses
        WHERE qAddress = '{address.street_address}'"""
    return query


def create_species_query(key) -> str:
    """Creates the sql query to retrieve the species and URL path for the species key. Returns the query as a string."""
    query = f"""
        SELECT qSpecies, urlPath
        FROM species
        WHERE "index" = {key}"""
    return query


def query_db(query: str):
    """General function to query the sqlite3 database at DB_LOCATION. Returns the results if they are found or None if there are none."""
    con = sqlite3.connect(DB_LOCATION)
    cur = con.cursor()
    result = cur.execute(query)

    if result:
        result = result.fetchall()

    con.close()
    return result


def get_species_keys(address: Address.Address) -> list[str] | None:
    """Queries the sqlite3 database for the species keys found at the given address.
    Returns a list of the keys found keys or None if none are found."""
    address_query = create_address_query(address)
    results = query_db(address_query)

    if results:
        keys = []
        for result in results:
            keys.append(str(result[0]))
        return keys
    else:
        return None


def get_species(key: str) -> tuple | None:
    """Queries the species table for the species matching the pass key. Returns the tuple (qSpecies, urlPath) if a species is found and returns None if none are found."""
    species_query = create_species_query(key)
    result = query_db(species_query)

    if result:
        result = result[0]  # should only be one species per key
        qSpecies = result[0]
        urlPaths = str(result[1])
        return qSpecies, urlPaths

    else:
        return None


def main(user_input: str, check_nearby: bool = True) -> pd.DataFrame | dict:
    """Main function that queries tree species from the given user_input. Returns a panda dataframe with the results."""
    # create an Address object from the given user input. Raises an exception if the input is not appropriate for the DB.
    try:
        query_address = Address.get_Address_for_query(user_input)
    except Address.NoCloseMatchError:
        raise Exception(f"Address {user_input} not found in San Francisco")
    except Address.AddressError as err:
        raise Exception(
            "Invalid address entered, ensure proper street address is given."
        )

    # namedtuple address_key to keep addresses and species keys together for later result formatting
    # SUGGEST may make more sense as dict(?)
    address_key = namedtuple("AddressKeys", ["address", "key"])
    address_keys = []
    species_keys = None

    # get all tree_ids at the query address
    species_keys = get_species_keys(query_address)

    # if there are trees there, make address/tree_id tuple (may change to dict)
    if not species_keys is None:
        for species_key in species_keys:
            address_keys.append(address_key(query_address.street_address, species_key))

    if species_keys is None and check_nearby:
        # if no trees at given address, will look next door (+2 or -2 street number i.e. 1470 and 1466 if given 1468)
        logging.warning(f"Couldn't find trees at given address, looking nearby...")
        steps = [-2, 4]

        for step in steps:
            query_address.street_number = int(query_address.street_number) + step
            species_keys = get_species_keys(query_address)
            if not species_keys is None:
                for species_key in species_keys:
                    address_keys.append(
                        address_key(query_address.street_address, species_key)
                    )

    if len(address_keys) == 0:
        raise Exception(f"Can't find any trees near entered street number")

    # create an empty results dataframe with number of rows as address_keys
    results = pd.DataFrame(
        columns=["qSpecies", "urlPath", "queried_address"],
        index=[i for i in range(len(address_keys))],
    )

    # save results to df
    for i, item in enumerate(address_keys):
        qSpecie, urlPath = get_species(item.key)
        results.loc[i] = {
            "qSpecies": qSpecie,
            "urlPath": urlPath,
            "queried_address": item.address.title(),
        }

    return results


def create_message(tree: pd.Series, number: int):
    # Moving to server repo
    scientific_name, common_name = split_qSpecies(tree.qSpecies)

    if not common_name is None:
        first_line = f"{common_name.title()} ({scientific_name.title()})"
    else:
        first_line = f"{scientific_name.title()}"

    if number > 1:
        first_line = first_line + f": {number}"

    if tree.urlPath != 0:
        second_line = f"https://selectree.calpoly.edu/tree-detail/{tree.urlPath}\n"
        return "\n".join([first_line, second_line])
    else:
        return first_line.append("\n")


def split_qSpecies(qSpecies: str) -> tuple:
    """Split qSpecies into a scientific name and common name. Common name will be returned as None if no common
    name is in the qSpecies field DEPRECATED."""
    split_species = qSpecies.split(" :: ")
    if len(split_species) == 1:
        scientific_name = split_species[0].replace(" ::", "")
        common_name = None
    else:
        scientific_name, common_name = split_species

    return scientific_name, common_name


def create_output_dict(results: pd.DataFrame) -> list[dict]:
    """Creates a formatted list of string messages from main()'s results df."""
    queried_addresses = results.queried_address.unique()
    tree_dict = {}

    for address in queried_addresses:
        grouped_results = (
            results.groupby(["qSpecies", "urlPath"])
            .count()
            .reset_index()
            .rename({"queried_address": "count"}, axis=1)
        )
        grouped_results[
            ["scientific_name", "common_name"]
        ] = grouped_results.qSpecies.str.split("::", expand=True).fillna("")
        grouped_results = (
            grouped_results.drop("qSpecies", axis=1)
            .loc[:, ["scientific_name", "common_name"]]
            .apply(lambda name: name.str.strip())
        )
        tree_dict.update({address: grouped_results.to_dict(orient="records")})
    return tree_dict


def format_output(results: pd.DataFrame) -> list[str]:
    """Creates a formatted list of string messages from main()'s results df. DEPRECATED USE"""
    queried_addresses = results.queried_address.unique()
    messages = list()

    for address in queried_addresses:
        address_line = f"Trees at {address}:\n"
        num_species = len(results.qSpecies.unique())

        if len(results) == 1:
            address_line = f"Tree at {address}:\n"
            messages.append(address_line + create_message(results.iloc[0], 1))

        elif num_species == 1:
            messages.append(
                address_line + create_message(results.iloc[0], len(results))
            )

        else:
            grouped_results = (
                results.groupby(["qSpecies", "urlPath"])
                .count()
                .reset_index()
                .rename({"queried_address": "count"}, axis=1)
            )
            first_message = messages.append(
                address_line
                + create_message(
                    grouped_results.iloc[0], grouped_results.loc[0, "count"]
                )
            )

            for tree in grouped_results[1:].itertuples():

                messages.append(create_message(tree, tree.count))

    return messages


def get_trees(user_input: str) -> list[str]:
    """Takes a string address from a user and returns a dictionary of the format:
    {address_1: [{
                common_name: str,
                scientific_name: str,
                urlPath: str,
                count: str
                },{}...]
     address_2: [{}, ...]
     }."""

    tree_df = main(user_input)
    return create_output_dict(tree_df)


if __name__ == "__main__":
    print(get_trees("1204 19th St"))
